from sklearn.base import BaseEstimator, TransformerMixin


class BasePreprocessor(BaseEstimator, TransformerMixin):

    def apply(self, elements):
        raise Exception("You have to implements apply")

    def fit(self, X, y=None):
        return self

    def _transform(self, X, func):
        if isinstance(X, dict):
            for k in X:
                X[k] = [func(row) for row in X[k]]
            return X
        elif isinstance(X, (list, tuple)):
            for i in range(len(X)):
                row = X[i]
                if len(row) > 0 and isinstance(row[0], (list, tuple)):
                    # row has multiple columns
                    _row = []
                    for column in row:
                        _row.append(func(column))
                    X[i] = _row
                else:
                    X[i] = func(row)
        return X

    def transform(self, X):
        return self._transform(X, self.apply)



import numpy as np
from itertools import compress


class Feed():

    def __init__(self, dataset, fields, field_adjuster=()):
        self.dataset = dataset
        self.fields = fields
        self.field_adjuster = field_adjuster
        if len(self.field_adjuster) == 0:
            self.field_adjuster = {}

    def full(self):
        batch_data = []
        for f in self.fields:
            adjuster = None
            if f in self.field_adjuster:
                adjuster = self.field_adjuster[f]
            batch_data.append(BatchData(self.dataset[f], adjuster))
        return batch_data

    def get_generator(self, batch_size, epoch=-1):
        sample_size = 0
        for f in self.fields:
            if sample_size == 0:
                sample_size = len(self.dataset[f])
            elif sample_size != len(self.dataset[f]):
                raise Exception("Data size does not match in dataset.")

        steps_per_epoch = sample_size // batch_size

        def generator():
            indices = np.arange(sample_size)
            count = 0
            _epoch = 0
            while True:
                if count > 0 and count % steps_per_epoch == 0:
                    _epoch += 1
                    count = 0
                    if epoch > 0 and _epoch >= epoch:
                        break
                batch = []
                candidates = np.zeros(sample_size)
                selected = np.random.choice(indices, size=batch_size,
                                            replace=False)
                candidates[selected] = 1
                for f in self.fields:
                    _selected = compress(self.dataset[f], candidates)
                    adjuster = None
                    if f in self.field_adjuster:
                        adjuster = self.field_adjuster[f]
                    batch.append(BatchData(_selected, adjuster))

                count += 1
                yield batch

        return steps_per_epoch, generator

    def batch_iter(self, batch_size, epoch=-1):
        steps_per_epoch, generator = self.get_generator(batch_size, epoch)
        for b in generator():
            yield b


class BatchData():

    def __init__(self, array, adjuster=None):
        self.array = array
        self.adjuster = adjuster

    def __call__(self):
        return np.array(self.array)

    def adjust(self, padding=-1, bos=False, eos=False, to_categorical=False):
        adjusted = [self.adjuster.adjust(x, padding, bos, eos)
                    for x in self.array]

        adjusted = np.array(adjusted)
        if to_categorical:
            return self.adjuster.to_categorical(adjusted)
        else:
            return adjusted

    def to_int_array(self):
        return np.array([int(i) for i in self.array])



import os
from pathlib import Path
import shutil
import json
import pandas as pd
from sklearn.externals import joblib
from chariot.feed import Feed
from chariot.transformer.adjuster import Adjuster
from chariot.resource.csv_file import CsvFile


class Dataset():

    def __init__(self, data_file, fields):
        self.data_file = data_file
        self.fields = fields

    @property
    def file_root(self):
        return os.path.dirname(self.data_file.path)

    def to_dataframe(self):
        if isinstance(self.data_file, CsvFile):
            delimiter = self.data_file.delimiter
            df = pd.read_csv(self.data_file.path, delimiter=delimiter,
                             header=None, names=self.fields)
            return df
        else:
            df = pd.read_table(self.data_file.path,
                               header=None, names=self.fields)
            return df

    def __call__(self, *fields):
        for result in self.fetch(target_fields=fields):
            yield result

    def get(self, *fields):
        dataset = {}

        if len(fields) == 0:
            _fields = self.fields
        elif isinstance(fields[0], (list, tuple)):
            _fields = fields[0]
        else:
            _fields = fields

        for c in _fields:
            dataset[c] = []

        for result in self.fetch(target_fields=_fields):
            for i, c in enumerate(_fields):
                dataset[c].append(result[i])

        return dataset

    def fetch(self, target_fields=(), progress=False):
        _target = target_fields
        if len(_target) == 0:
            _target = self.fields  # Select all fields

        first = True
        for line in self.data_file.fetch(progress):
            if first and isinstance(line, (list, tuple)):
                # change field name to index
                _target = [i for i, c in enumerate(self.fields)
                           if c in _target]

            if isinstance(line, (dict, list, tuple)):
                result = [line[c] for c in _target]
            else:
                result = line

            first = False
            yield result

    def save_transformed(self, transform_name, field_transformers):
        target_fields = [c for c in self.fields if c in field_transformers]
        dataset = self.get(target_fields)

        for f in field_transformers:
            if field_transformers[f]:
                dataset[f] = field_transformers[f].transform(dataset[f])

        transformed = TransformedDataset.create(transform_name, dataset,
                                                field_transformers,
                                                self.data_file)
        return transformed

    def to_feed(self, field_transformers=(),
                apply_transformer=True):
        target_fields = [c for c in self.fields if c in field_transformers]
        dataset = self.get(target_fields)

        field_adjuster = {}
        for f in dataset:
            aj = None
            if f in field_transformers:
                if field_transformers[f] is None:
                    continue
                t = field_transformers[f]
                if apply_transformer:
                    dataset[f] = t.transform(dataset[f])
                if t.indexer is not None:
                    indexer = t.indexer
                    aj = Adjuster(len(indexer.vocab), indexer.pad,
                                  indexer.unk, indexer.bos, indexer.eos)

            if aj is not None:
                field_adjuster[f] = aj

        return Feed(dataset, target_fields, field_adjuster)


class TransformedDataset(Dataset):

    def __init__(self, data_file, original_path, field_transformers):
        fields = list(field_transformers.keys())
        super().__init__(data_file, fields)
        self.original_path = original_path
        self.field_transformers = field_transformers

    @classmethod
    def create(cls, transform_name, data, field_transformers, source_file):
        original_path = source_file.path
        transed = source_file.convert(data_dir_to="processed",
                                      add_attribute=transform_name)
        # Make transformed dir
        root = os.path.join(os.path.dirname(transed.path), transed.base_name)
        root = Path(root)

        if root.exists():
            shutil.rmtree(root)
        root.mkdir()

        # Save transformers
        fields = []
        for f in field_transformers:
            t = field_transformers[f]
            if t is not None:
                joblib.dump(t, root.joinpath("{}.pkl".format(f)))
            fields.append(f)

        # Save Data
        transed = transed.convert(data_dir_to="processed/" + root.name)
        field_data = [data[c] for c in field_transformers]
        with open(transed.path, mode="w", encoding=transed.encoding) as f:
            for elements in zip(*field_data):
                strs = []
                for e in elements:
                    if isinstance(e, (list, tuple)):
                        strs.append(" ".join([str(_e) for _e in e]))
                    else:
                        strs.append(str(e))

                line = ""
                if len(strs) > 1:
                    line = transed.delimiter.join(strs)
                else:
                    line = strs[0]

                f.write(line + "\n")

        # Make metadata
        meta_data = {
            "original": os.path.abspath(original_path),
            "fields": fields
        }

        with root.joinpath("meta.json").open(
                mode="w", encoding="utf-8") as f:
                json.dump(meta_data, f)

        instance = cls(transed, original_path, field_transformers)
        return instance

    @classmethod
    def load(cls, source_file, transform_name):
        transed = source_file.convert(data_dir_to="processed",
                                      add_attribute=transform_name)
        root = os.path.join(os.path.dirname(transed.path),
                            transed.base_name)
        root = Path(root)
        transed_file = transed.convert(data_dir_to="processed/" + root.name)

        with root.joinpath("meta.json").open(encoding="utf-8") as f:
            meta = json.load(f)

        original_path = meta["original"]
        fields = meta["fields"]

        field_transformers = {}
        for f in fields:
            if root.joinpath(f + ".pkl").exists():
                t = joblib.load(root.joinpath(f + ".pkl"))
                field_transformers[f] = t
            else:
                field_transformers[f] = None

        instance = cls(transed_file, original_path, field_transformers)
        return instance

    def fetch(self, target_fields=(), progress=False):
        _target = target_fields
        if len(_target) == 0:
            _target = self.fields

        for result in super().fetch(target_fields, progress):
            _result = []
            for r, t in zip(result, _target):
                if t in self.field_transformers:
                    # Space separated index string to int array.
                    _result.append([int(i) for i in r.split(" ")])
                else:
                    _result.append(r)
            yield _result



class Token():

    def __init__(self, token, token_type="spaCy"):
        self._token = token
        self.token_type = token_type

    @property
    def is_spacy(self):
        return self.token_type == "spaCy"

    @property
    def is_ja(self):
        return self.token_type == "ja"

    @property
    def surface(self):
        if self.token_type == "spaCy":
            return self._token.text
        elif self.token_type == "ja":
            return self._token.surface

    @property
    def base_form(self):
        if self.token_type == "spaCy":
            return self._token.lemma_
        elif self.token_type == "ja":
            return self._token.base_form

    @property
    def pos(self):
        if self.token_type == "spaCy":
            return self._token.pos_
        elif self.token_type == "ja":
            return self._token.part_of_speech[0]

    @property
    def tag(self):
        if self.token_type == "spaCy":
            return self._token.tag_
        elif self.token_type == "ja":
            return self._token.part_of_speech[1]

    def __repr__(self):
        return "<{}:{}>".format(self.surface, self.pos)



import numpy as np


class Adjuster():

    def __init__(self, vocab_size=0, padding=-1, unknown=-1,
                 begin_of_seq=-1, end_of_seq=-1):
        self.vocab_size = vocab_size
        self.padding = padding
        self.unknown = unknown
        self.begin_of_seq = begin_of_seq
        self.end_of_seq = end_of_seq

    def adjust(self, indices, padding=-1, bos=False, eos=False):
        if bos:
            indices = [self.begin_of_seq] + indices
        if eos:
            indices = indices + [self.end_of_seq]

        if padding > 0:
            pad = self.padding
            if len(indices) < padding:
                pad_size = padding - len(indices)
                indices = indices + [pad] * pad_size
            elif len(indices) > padding:
                indices = indices[:padding]

        return indices

    def to_categorical(self, array):
        y = np.array(array, dtype="int")
        input_shape = y.shape
        y = y.ravel()
        n = y.shape[0]
        categorical = np.zeros((n, self.vocab_size), dtype=np.float32)
        categorical[np.arange(n), y] = 1
        output_shape = input_shape + (self.vocab_size,)
        categorical = np.reshape(categorical, output_shape)
        return categorical



from collections import Counter
import numbers
from chariot.transformer.base_preprocessor import BasePreprocessor
from chariot.transformer.tokenizer.token import Token


class Indexer(BasePreprocessor):

    def __init__(self, padding="__PAD__", unknown="__UNK__",
                 begin_of_seq="__BOS__", end_of_seq="__EOS__",
                 min_df=1, size=-1):
        self._vocab = []
        self._padding = padding
        self._unknown = unknown
        self._begin_of_seq = begin_of_seq
        self._end_of_seq = end_of_seq
        self.min_df = min_df
        self.size = size

    @property
    def vocab(self):
        return self._vocab

    @property
    def unk(self):
        return self._vocab.index(self._unknown)

    @property
    def pad(self):
        return self._vocab.index(self._padding)

    @property
    def bos(self):
        return self._vocab.index(self._begin_of_seq)

    @property
    def eos(self):
        return self._vocab.index(self._end_of_seq)

    def token_to_words(self, tokens):
        words = [t if isinstance(t, str) else t.surface for t in tokens]
        return [w.strip() for w in words]

    def apply(self, words):
        _words = self.token_to_words(words)
        indices = []
        for w in _words:
            if w in self.vocab:
                indices.append(self.vocab.index(w))
            else:
                indices.append(self.unk)
        return indices

    def inverse(self, indices, exclude_padding=True):
        vocab = self.vocab
        pad = self.pad
        # Make text exclude padding
        words = [vocab[i] for i in indices]
        if exclude_padding:
            words = [vocab[i] for i in indices if i != pad]
        return words

    def transform(self, X):
        if len(self._vocab) == 0:
            raise Exception("Vocabulary has not made yet. Plase execute fit.")
        return super().transform(X)

    def inverse_transform(self, X):
        if len(self._vocab) == 0:
            raise Exception("Vocabulary has not made yet. Plase execute fit.")
        return self._transform(X, self.inverse)

    def fit(self, X, y=None):
        vocab = Counter()
        dataset = []

        # X should be list of tokens or dict
        if isinstance(X, dict):
            for columns in X:
                for tokens in X[columns]:
                    vocab.update(self.token_to_words(tokens))
        else:
            for row in X:
                if len(row) > 0 and isinstance(row[0], (list, tuple)):
                    # row has multiple columns
                    for tokens in row:
                        vocab.update(self.token_to_words(tokens))
                else:
                    vocab.update(self.token_to_words(row))

        reserved = [self._padding, self._unknown,
                    self._begin_of_seq, self._end_of_seq]
        reserved = [r for r in reserved if r]

        selected = reserved
        if self.min_df >= 0:
            md = self.min_df
            min_count = (md if isinstance(md, numbers.Integral)
                         else md * len(dataset))
            for term, count in vocab.most_common():
                if count <= min_count:
                    continue
                else:
                    selected.append(term)
        elif self.size > 0:
            for term, count in vocab.most_common():
                if len(selected) < self.size:
                    selected.append(term)
                else:
                    break

        self._vocab = selected



import os
import re
from zipfile import ZipFile
from chariot.resource.data_file import DataFile
from chariot.resource.csv_file import CsvFile


class Storage():

    def __init__(self, root):
        """Provide the access to the data folder.
        The data folder structure should follow the Cookiecutter Data Science
        manner.
        https://drivendata.github.io/cookiecutter-data-science/

        ROOT
         └── data
              ├── external     <- Data from third party sources.
              ├── interim      <- Intermediate data that has been transformed.
              ├── processed    <- The final, canonical data sets for modeling.
              └── raw          <- The original, immutable data dump.

        Args:
            root: path to root directory (parent directory of data, models).
        """
        self.root = root

    def path(self, target=""):
        return os.path.join(self.root, "data/{}".format(target))

    def file(self, target, encoding="utf-8", delimiter=",", has_header=False):
        path = self.path(target)
        _, ext = os.path.splitext(path)
        if ext in [".csv", ".tsv"]:
            return CsvFile(path, encoding, delimiter, has_header)
        else:
            return DataFile(path, encoding)

    def chazutsu(self, path_or_resource, columns=None, target="",
                 separator="\t", pattern=()):

        from chariot.resource.chazutsu_resource import ChazutsuResource
        return ChazutsuResource(path_or_resource, columns,
                                target, separator, pattern)

    def _to_snake(self, name):
        _name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
        _name = re.sub("([a-z0-9])([A-Z])", r"\1_\2", _name).lower()
        return _name

    def expand_zip(self, path):
        file_name = os.path.basename(path)
        file_root, ext = os.path.splitext(file_name)
        location = os.path.dirname(path)
        archive_dir = os.path.join(location, file_root)
        if os.path.exists(archive_dir):
            print("The file {} is already expanded".format(
                    os.path.abspath(path)))

        with ZipFile(path) as zip:
            zip.extractall(location)





import csv
from tqdm import tqdm
from chariot.resource.data_file import DataFile


class CsvFile(DataFile):

    def __init__(self, path, encoding="utf-8", delimiter=",",
                 has_header=False):
        super().__init__(path, encoding)
        self.delimiter = delimiter
        self.has_header = has_header
        self.header = []

    def to_dataset(self, fields=()):
        from chariot.dataset import Dataset
        fields = list(fields)
        if len(fields) == 0:
            fields = self.header
        return Dataset(self, fields)

    def fetch(self, progress=False):
        done_header = not self.has_header
        total_count = 0
        if progress:
            total_count = self.get_line_count()

        with open(self.path, encoding=self.encoding) as f:
            reader = csv.reader(f, delimiter=self.delimiter)
            if progress:
                reader = tqdm(reader, total=total_count)

            for row in reader:
                if self.has_header and not done_header:
                    if len(self.header) == 0:
                        self.header = row
                    done_header = True
                else:
                    if len(self.header) > 0:
                        element = {}
                        for name, item in zip(self.header, row):
                            element[name] = item
                        yield element
                    else:
                        yield row

    def write(self, elements_iter):
        with open(self.path, mode="wb", encoding=self.encoding) as f:
            writer = csv.writer(f, delimiter=self.delimiter)
            for elements in elements_iter:
                writer.writerow(elements)



from chariot.resource.csv_file import CsvFile
from chariot.dataset import Dataset
from chazutsu.datasets.framework.resource import Resource


class ChazutsuResource():

    def __init__(self, resource, columns=None, target="",
                 separator="\t", pattern=()):
        if isinstance(resource, Resource):
            self._resource = resource
        else:
            r = Resource(resource, columns, target, separator,
                         pattern)
            self._resource = r

    @property
    def data_file(self):
        return self._get_csv_file("data")

    @property
    def train_file(self):
        return self._get_csv_file("train")

    @property
    def test_file(self):
        return self._get_csv_file("test")

    @property
    def valid_file(self):
        return self._get_csv_file("valid")

    @property
    def sample_file(self):
        return self._get_csv_file("sample")

    @property
    def dataset(self):
        return self._get_dataset("data")

    @property
    def train_dataset(self):
        return self._get_dataset("train")

    @property
    def test_dataset(self):
        return self._get_dataset("test")

    @property
    def valid_dataset(self):
        return self._get_dataset("valid")

    @property
    def sample_dataset(self):
        return self._get_dataset("sample")

    def _get_csv_file(self, kind):
        path = self._resource._get_prop(kind)
        separator = self._resource.separator
        file_instance = CsvFile(path, delimiter=separator, has_header=False)
        if self._resource.columns is None:
            file_instance.header = self._resource.columns
        return file_instance

    def _get_dataset(self, kind):
        data_file = self._get_csv_file(kind)
        fields = data_file.header
        return Dataset(data_file, fields)



from chariot.tokenizer import Tokenizer
from chariot.preprocessor.text import TextFilter, TextNormalizer
from chariot.preprocessor.token import TokenFilter, TokenNormalizer


class Parser():

    def __init__(self, tokenizer_or_lang,
                 text_preprocessors=(), token_preprocessors=()):
        self.text_preprocessors = text_preprocessors
        self.tokenizer = tokenizer_or_lang
        if isinstance(self.tokenizer, str):
            self.tokenizer = Tokenizer(tokenizer_or_lang)
        self.token_preprocessors = token_preprocessors

    @classmethod
    def create(cls, steps):
        text_ps = []
        token_ps = []
        tokenizer = None

        for p in steps:
            if isinstance(p, (TextFilter, TextNormalizer)):
                text_ps.append(p)
            elif isinstance(p, Tokenizer):
                tokenizer = p
            elif isinstance(p, (TokenFilter, TokenNormalizer)):
                token_ps.append(p)

        if tokenizer is None:
            raise Exception("Tokenizer should be included to parser.")

        return cls(tokenizer, text_ps, token_ps)

    def parse(self, text, return_surface=False):
        _text = text
        for tp in self.text_preprocessors:
            _text = tp.apply(_text)

        tokens = self.tokenizer.tokenize(_text)

        for tp in self.token_preprocessors:
            tokens = tp.apply(tokens)

        if return_surface:
            return [t.surface for t in tokens]
        else:
            return tokens
